/* -*- mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*- */
/*
 *  Main authors:
 *     Christian Schulte <schulte@gecode.org>
 *
 *  Copyright:
 *     Christian Schulte, 2005
 *
 *  Last modified:
 *     $Date: 2009-04-02 20:58:46 +0200 (Thu, 02 Apr 2009) $ by $Author: schulte $
 *     $Revision: 8649 $
 *
 *  This file is part of Gecode, the generic constraint
 *  development environment:
 *     http://www.gecode.org
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <gecode/driver.hh>
#include <gecode/int.hh>

using namespace Gecode;

/**
 * \brief %Example: Stress distinct propagator
 *
 * Performs propagation on a pathological example. The example is taken
 * from: Jean-Francois Puget, A fast algorithm for the bound consistency
 * of alldiff constraints, AAAI-98, pages 359-366, July 1998,
 * Madison, WI, USA.
 *
 * \ingroup ExStress
 */
class StressDistinct : public Script {
protected:
  /// Variables
  IntVarArray x;
public:
  /// The actual problem
  StressDistinct(const SizeOptions& opt)
    : x(*this,opt.size()+1,0,opt.size()) {
    distinct(*this, x, opt.icl());
    for (unsigned int i=0; i<opt.size(); i++)
      rel(*this, x[i], IRT_LQ, i);
  }
  /// Constructor for cloning \a s
  StressDistinct(bool share, StressDistinct& s) : Script(share,s) {
    x.update(*this, share, s.x);
  }
  /// Perform copying during cloning
  virtual Space*
  copy(bool share) {
    return new StressDistinct(share,*this);
  }
  /// Print solution
  virtual void
  print(std::ostream& os) const {
    os << "\tx[" << x.size() << "] = " << x << std::endl;
  }
};

/** \brief Main-function
 *  \relates StressDistinct
 */
int
main(int argc, char* argv[]) {
  SizeOptions opt("StressDistinct");
  opt.size(1000);
  opt.parse(argc,argv);
  Script::run<StressDistinct,DFS,SizeOptions>(opt);
  return 0;
}

// STATISTICS: example-any
